# -*- coding: utf-8 -*-
"""volume discretization for torchgdm
"""
from .pola import StructDiscretized2D, StructDiscretizedSquare2D
from . import pola
from . import geometries

from .geometries import discretizer_square

from .geometries import square
from .geometries import rectangle
from .geometries import circle
from .geometries import split_ring
from .geometries import triangle_equilateral
