# encoding=utf-8
"""environment for 2D structures in a homogeneous environment.
"""
from . import inc_fields

from .dyads import EnvHomogeneous2D

from .inc_fields import NullField
from .inc_fields import PlaneWave
from .inc_fields import ElectricLineDipole
from .inc_fields import MagneticLineDipole